@DATABASE "GTEvents.mod"
(*
@NODE MAIN "GTEvents" *)
MODULE GTEvents;
(*
     $RCSfile: GTEvents.mod $
  Description: Implementation of GadToolsPort class

   Created by: fjc (Frank Copeland)
    $Revision: 1.5 $
      $Author: fjc $
        $Date: 1994/09/03 16:25:29 $

  Copyright  1994, Frank Copeland.
  This file is part of the Oberon-A Library.
  See Oberon-A.doc for conditions of use and distribution.
  ________________________________________________________________________

  Module GTEvents extends the classes defined in Module Events to provide
  the event management required by the GadTools library introduced in
  Kickstart 2.0.

  @{" Overview " link Overview}  Overview of GadTools event handling

  Class descriptions

  @{" GadToolsPort " link GadToolsPort}  Description of the GadToolsPort Class

  Other

  @{" Switches       " link Switches}  Global compiler switches
  @{" Imports        " link Imports}  Imported modules
  @{" Initialisation " link InitModule}  Module initialisation
  @{" Terminology    " link Terminology.doc/MAIN}  Definition of terms
  @{" History        " link History}  Revision history
@ENDNODE

@NODE Overview "Overview of GadTools event handling"
  The GadTools library was introduced in Kickstart 2.0 to provide high
  level user-interface objects that previously had to be built by hand
  using Intuition's lower level facilities.  Notification of events
  involving these objects is via the existing IDCMP message ports.
  However, GadTools objects are often complex, composite collections of
  related Gadgets, requiring a lot of work to manage.  To spare
  programmers this task, events related to these Gadgets are processed and
  filtered by the GadTools library, which ensures that programmers only
  receive notification of events that are of interest to them.

  To perform this filtering, GadTools library requires that programmers
  replace calls to Exec library's GetMsg() and ReplyMsg() functions with
  calls to GadTools' own GetIMsg() and ReplyIMsg() functions.  These do
  the actual task of filtering events for GadTools objects, but pass
  through other events unchanged.

  The GadToolsPort class is simply an extension of the IdcmpPort class
  which overrides the methods that rely on the Exec functions and replaces
  them with methods that use the GadTools functions.
@ENDNODE

@NODE History "Revision History"
  $Revision: 1.5 $
      $Date: 1994/09/03 16:25:29 $
  ________________________________________________________________________

  $Log: GTEvents.mod $
  Revision 1.5  1994/09/03  16:25:29  fjc
  *** empty log message ***

  Revision 1.4  1994/08/08  16:19:59  fjc
  Release 1.4

  Revision 1.3  1994/06/14  02:06:07  fjc
  - Updated for release

  Revision 1.2  1994/06/04  15:50:17  fjc
  - Changed to use new Amiga interface

  Revision 1.1  1994/05/12  19:56:40  fjc
  - Prepared for release


@ENDNODE

@NODE Switches "Global Switches"

  $C= CaseChk       $I= IndexChk  $L+ LongAdr   $N- NilChk
  $P= PortableCode  $R= RangeChk  $S= StackChk  $T= TypeChk
  $V= OvflChk       $Z= ZeroVars

  NIL checking is switched off and replaced with ASSERTs at the relevant
  points in the code.

@ENDNODE
*)

(*
@NODE Imports "Imported Modules"*)

IMPORT
  E := Exec, I := Intuition, GT := GadTools, Ev := Events, SYS := SYSTEM;

(*
  @{" Exec      " link OBERON-A:Amiga/Exec.mod/MAIN}  Exec library
  @{" Intuition " link OBERON-A:Amiga/Intuition.mod/MAIN}  Intuition library
  @{" GadTools  " link OBERON-A:Amiga/Libraries/GadTools.mod/MAIN}  GadTools library
  @{" Events    " link Events.mod/MAIN}  Module Events
@ENDNODE *)

(*
@NODE GadToolsPort "Description of GadToolsPort"
  A GadToolsPort is basically an @{"IdcmpPort" link Events.mod/IdcmpPort} that has been modified to use
  the GadTools GetIMsg() and ReplyIMsg() functions.

  ________________________________________________________________________

  @{" Declarations " link GadToolsPortDecls}  Constant and type declarations
  @{" Methods      " link GadToolsPortMethods}  GadToolsPort methods
  @{" Usage        " link GadToolsPortUsage}  Using GadToolsPort objects
@ENDNODE

@NODE GadToolsPortUsage "Using GadToolsPort"
  A GadToolsPort is used in exactly the same way as an IdcmpPort.  Simply
  replace any variables of type IdcmpPort with variables of type
  GadToolsPort.  The only other requirement is to use ReplyIMsg() instead
  of ReplyMsg() in handler procedures that consume messages:

  (*------------------------------------*)
  PROCEDURE* HandleCloseWindow
    ( ip : IdcmpPort; msg : I.IntuiMessagePtr )
    : INTEGER;

  BEGIN (* HandleCloseWindow *)
    GT.base.ReplyIMsg (msg);
    RETURN Stop
  END HandleCloseWindow;

@ENDNODE *)

(*------------------------------------------------------------------------
@NODE GadToolsPortDecls "GadToolsPort Declarations" *)
TYPE

  GadToolsPort *= POINTER TO GadToolsPortRec;
  GadToolsPortRec *= RECORD (Ev.IdcmpPortRec) END;

(*
@ENDNODE *)

(*------------------------------------------------------------------------
@NODE GadToolsPortMethods "GadToolsPort Methods"
  OVERRIDDEN METHODS

  Event handling

    PROCEDURE ( gtp : GadToolsPort) @{"HandleSig" link HandleSig} * () : INTEGER;
    - Removes and processes messages queued at the object's MessagePort.

  Message Port maintenance

    PROCEDURE ( gtp : GadToolsPort ) @{"FlushPort" link FlushPort}* ();
    - Flushes any pending messages from the object's MessagePort.

@ENDNODE *)

(*
@NODE HandleSig "HandleSig()" *)
PROCEDURE (gtp : GadToolsPort) HandleSig * () : INTEGER;
(*
  This procedure implements the behaviour required by receiving the signal
  associated with a MsgPort.  The actual handling of the messages is
  delegated to gtp.HandleMsg().
*)

  VAR result : INTEGER; msg : I.IntuiMessagePtr;

BEGIN (* HandleSig *)
  result := Ev.Pass; (* Default *)

  (* Loop until all messages queued at the port are dealt with *)
  ASSERT (gtp.port # NIL, 132);
  LOOP
    (* Dequeue the next message, quit if there is none *)
    msg := GT.base.GetIMsg (gtp.port);
    IF msg = NIL THEN EXIT END;

    (* Despatch to the message handler *)
    result := gtp.HandleMsg (msg);

    (* Process the return code *)
    IF result = Ev.Pass THEN GT.base.ReplyIMsg (msg) END;
    IF result > Ev.Continue THEN EXIT END
  END;
  RETURN result
END HandleSig;
(*
@ENDNODE *)

(*
@NODE FlushPort "FlushPort()" *)
PROCEDURE (gtp : GadToolsPort) FlushPort * ();
(*
  Gets and replies to all messages queued for the handler's message
  port.  It is called inside an Exec.Forbid()/Exec.Permit() pair to
  prevent more messages from arriving at the port while it is being
  flushed.
*)

  VAR msg : I.IntuiMessagePtr;

BEGIN (* FlushPort *)
  ASSERT (gtp.port # NIL, 132);
  E.base.Forbid ();
    LOOP
      msg := GT.base.GetIMsg (gtp.port);
      IF msg = NIL THEN EXIT END;
      GT.base.ReplyIMsg (msg)
    END;
  E.base.Permit ()
END FlushPort;
(*
@ENDNODE *)

(*
@NODE InitModule "Module Initialisation" *)

(* No initialisation required *)

END GTEvents.
(*
@ENDNODE *)

